# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    from importlib_metadata import (  # type: ignore[import-not-found, no-redef]
        PackageNotFoundError,
        version,
    )

UNINSTALLED = "uninstalled"

try:
    VERSION = version("vuepy-core")
except PackageNotFoundError:
    VERSION = UNINSTALLED


def get_semver_version(_version: str) -> str:
    """
    https://semver.org/
    1.0.0-alpha < 1.0.0-alpha.1 < 1.0.0-alpha.beta < 1.0.0-beta < 1.0.0-beta.2 < 1.0.0-beta.11 < 1.0.0-rc.1 < 1.0.0

    :param version:
    :return:
    """
    if _version == UNINSTALLED:
        return _version

    split = _version.split(".", maxsplit=2)
    patch = split[2] if len(split) == 3 else ''
    is_pre_release = "a" in patch or "b" in patch
    if is_pre_release:
        return ".".join(split)

    return "~" + ".".join([split[0], split[1], "*"])


SEMVER_VERSION = get_semver_version(VERSION)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import dataclasses
import pathlib
import re
from html.parser import HTMLParser
from typing import List
from typing import Tuple

from vuepy import log
from vuepy.compiler_sfc.script_compiler import ScriptCompiler

logger = log.getLogger()


@dataclasses.dataclass
class SFCTag:
    """
    start_pos: [row=1,col=0]
    |
    <tag>..</tag>
           |
           end_pos
    """
    name: str
    attrs: dict
    start_pos: Tuple[int, int]
    end_pos: Tuple[int, int] = None
    inner_html: str = ''

    def to_dict(self):
        return {
            'name': self.name,
            'attrs': self.attrs,
            'start_pos': self.start_pos,
            'end_pos': self.end_pos,
            'inner_html': self.inner_html,
        }


class SFCParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.tag_stack = []
        self.level1_tags = []
        self.sfc_tags: List[SFCTag] = []

    def handle_starttag(self, tag, attrs):
        if not self.tag_stack:
            self.level1_tags.append(SFCTag(tag, dict(attrs), self.getpos()))
        self.tag_stack.append(tag)

    def handle_endtag(self, tag):
        if self.tag_stack:
            self.tag_stack.pop()

        if not self.tag_stack and self.level1_tags:
            sfc_tag = self.level1_tags.pop()
            if sfc_tag.name != tag:
                row_s, _ = sfc_tag.start_pos
                row_e, _ = self.getpos()
                msg = f"tag <{sfc_tag.name}> at {sfc_tag.start_pos} mismatch: " \
                      f"expected </{sfc_tag.name}>, got </{tag}> at {self.getpos()}.\n" \
                      f"-> {row_s} {self.html_lines[row_s - 1]}...\n" \
                      f"-> {row_e} {self.html_lines[row_e - 1]}"
                raise ValueError(msg)

            sfc_tag.end_pos = self.getpos()
            # todo
            s = self.vue_html.find('>', self.calc_idx_by_pos(sfc_tag.start_pos)) + 1
            e = self.calc_idx_by_pos(sfc_tag.end_pos)
            sfc_tag.inner_html = self.vue_html[s:e]

            self.sfc_tags.append(sfc_tag)

    def calc_idx_by_pos(self, pos):
        row, col = pos
        return self.html_lines_len_sum[row] + col

    def parse(self, vue_html) -> List[SFCTag]:
        self.vue_html = vue_html
        self.html_lines = vue_html.splitlines(keepends=True)
        html_lines_len = [len(i) for i in self.html_lines]
        self.html_lines_len_sum = [None, 0]
        acc = 0
        for i, line_len in enumerate(html_lines_len):
            acc += html_lines_len[i]
            self.html_lines_len_sum.append(acc)

        self.tag_stack = []
        self.level1_tags = []
        self.feed(vue_html)
        # handle last </tag>
        if self.level1_tags:
            sfc_tag = self.level1_tags.pop()
            row, col = sfc_tag.start_pos
            msg = f"tag <{sfc_tag.name}> at {sfc_tag.start_pos} not closed.\n" \
                  f"-> {row} {self.html_lines[row-1]}"
            raise ValueError(msg)
        return self.sfc_tags


@dataclasses.dataclass
class SFCMetadata:
    file: pathlib.Path
    content: str
    template: str
    script_src: str
    script_py: str

    @classmethod
    def load(cls, sfc_file):
        sfc_file = pathlib.Path(sfc_file)
        with open(sfc_file) as f:
            raw_content = f.read()

        return cls.loads(raw_content, sfc_file)

    @classmethod
    def loads(cls, sfc_content, file_path='__tmp_for_str.vue') -> SFCMetadata:
        sfc_tags = SFCParser().parse(sfc_content)

        script_src_tag_attrs = {}
        script_py_tag: SFCTag = None
        template_tag = None
        for tag in sfc_tags:
            if tag.name == 'template':
                if template_tag:
                    raise ValueError("template tag duplicate")
                template_tag = tag
            elif tag.name == 'script':
                if 'src' in tag.attrs:
                    script_src_tag_attrs = tag.attrs
                elif tag.attrs.get('lang', '').lower() == 'py':
                    script_py_tag = tag

        if template_tag is None:
            raise ValueError(f"can't find <template> in {file_path}")

        instance = cls(
            file=pathlib.Path(file_path),
            content=sfc_content,
            template=template_tag.inner_html,
            script_src=script_src_tag_attrs.get("src"),
            script_py=script_py_tag and script_py_tag.inner_html,
        )
        if instance.script_py and instance.script_src:
            raise ValueError(
                "Syntax error: script lang and script src cannot exist at the same time"
            )

        return instance

    @staticmethod
    def get_block_content(tag, content):
        blocks = re.findall(f'<{tag}>(.*)</{tag}>', content, flags=re.S | re.I)
        return blocks

    @staticmethod
    def get_script_src(content):
        match = re.search(r"<script (.*?)src=(['\"])(?P<src>.*?)\2></script>", content)
        return match.group('src') if match else None

    @staticmethod
    def get_script_py(content):
        match = re.search(
            fr'<script\s+lang=(["\'])py\1\s*>(?P<content>.*)?</script>',
            content,
            flags=re.S | re.I
        )

        if not match:
            logger.debug("can't find <script lang=py>")
            return None

        return match.group('content')


def parse(sfc_file, raw_content=False) -> SFCMetadata:
    """
    Parse a sfc file.

    :param sfc_file: The path to the sfc file.
    :param raw_content: If True, the sfc file is parsed as a string.
    :return: The parsed sfc file metadata.
    """
    return SFCMetadata.loads(sfc_file) if raw_content else SFCMetadata.load(sfc_file)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import re
from typing import Any
from typing import Callable
from typing import List
from typing import Tuple

import ipywidgets as widgets

from vuepy import log
from vuepy.compiler_core.ast import NodeAst
from vuepy.compiler_core.ast import VForAst
from vuepy.compiler_core.ast import VueCompAst
from vuepy.compiler_core.component_expr import VueCompExpr
from vuepy.compiler_core.utils import VueCompNamespace
from vuepy.reactivity.reactive import to_raw
from vuepy.reactivity.watch import WatchOptions
from vuepy.reactivity.watch import watch

logger = log.getLogger()


def v_for_stack_to_iter(stack: List[VForAst], fn: VForIterFn, ns: dict,
                        for_block_stack_idxs: Tuple[int] = (),
                        for_block_stack_vars: dict = None) -> List[Any]:
    """
    s1 = [1,2]
    s2 = [3,4,5]
    s3 = [6,7]
    for_stack = [s1, s2, s3]
    --
    [ [ [ fn((0, 0, 0), (1, 3, 6)),
          fn((0, 0, 1), (1, 3, 7)) ],
        [ fn((0, 1, 0), (1, 4, 6)),
          fn((0, 1, 1), (1, 4, 7)) ],
        [ fn((0, 2, 0), (1, 5, 6)),
          fn((0, 2, 1), (1, 5, 7)) ]],
      [ [ fn((1, 0, 0), (2, 3, 6)),
          fn((1, 0, 1), (2, 3, 7)) ],
        [ fn((1, 1, 0), (2, 4, 6)),
          fn((1, 1, 1), (2, 4, 7)) ],
        [ fn((1, 2, 0), (2, 5, 6)),
          fn((1, 2, 1), (2, 5, 7)) ]]]

    :param stack:
    :param fn:
    :param ns:
    :param for_block_stack_idxs:
    :param for_block_stack_vars:
    :return:
    """
    if not stack:
        return []

    for_block_stack_vars = for_block_stack_vars or {}
    v_for_ast = stack[0]
    if len(stack) == 1:
        ret = []
        with VForBLockScope(v_for_ast, ns) as for_block_scope:
            for i, val in enumerate(for_block_scope):
                _idxs = (*for_block_stack_idxs, i)
                _vars = {**for_block_stack_vars, **for_block_scope.for_vars}
                ret.append(fn(_idxs, _vars, v_for_ast))
        return ret

    ret = []
    with VForBLockScope(v_for_ast, ns) as for_block_scope:
        for i, val in enumerate(for_block_scope):
            _idxs = (*for_block_stack_idxs, i)
            _vars = {**for_block_stack_vars, **for_block_scope.for_vars}
            ret.append(v_for_stack_to_iter(stack[1:], fn, for_block_scope.ns, _idxs, _vars))

    return ret


class VForBLockScope:
    def __init__(self, v_for_ast: VForAst, ns):
        self.ns = ns
        self.v_for_ast = v_for_ast

        self.iter_exp = None
        self._iter = None
        self.target = None
        self.idx = 0
        self.vars_bak = {}
        self.for_vars = {}

    def __enter__(self):
        iter_exp = self.v_for_ast.iter
        if iter_exp in self.ns:
            self.iter_exp = self.ns[iter_exp]
        else:
            self.iter_exp = eval(iter_exp, {}, self.ns)
        self._iter = iter(self.iter_exp)

        target_var = self.v_for_ast.target
        if target_var in self.ns:
            self.vars_bak[target_var] = self.ns[target_var]

        idx_var = self.v_for_ast.idx
        if idx_var in self.ns:
            self.vars_bak[idx_var] = self.ns[idx_var]

        self.for_vars[iter_exp] = self.iter_exp

        return self

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        # set target
        target_exp = self.v_for_ast.target
        target = next(self._iter)
        self.for_vars[target_exp] = target
        self.ns[target_exp] = target

        if self.v_for_ast.idx is not None:
            # set idx
            idx_exp = self.v_for_ast.idx
            self.for_vars[idx_exp] = self.idx
            self.ns[idx_exp] = self.idx

        self.idx += 1

        return target

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.ns.update(self.vars_bak)
        self.vars_bak = {}
        self.for_vars = {}


VForIterFn = Callable[[Tuple[int], dict, VForAst], Any]


class VueHtmlTemplateRender:
    DELIMITERS_PATTERN = r"\{\{\s*(.*?)\s*\}\}"

    @staticmethod
    def _replace(ns: VueCompNamespace, for_idx):
        def warp(match):
            expr_str = match.group(1)
            expr_ast = VueCompExpr.parse(expr_str)
            # TODO html可以设置value，按需更新
            _value = expr_ast.eval(ns)
            return str(_value)

        return warp

    @classmethod
    def should_render(cls, template):
        return bool(re.search(cls.DELIMITERS_PATTERN, template))

    @classmethod
    def render(cls, template, ns: VueCompNamespace, for_idx=-1):
        result = re.sub(cls.DELIMITERS_PATTERN, cls._replace(ns, for_idx), template)
        return result


class VueHtmlCompCodeGen:
    @classmethod
    def gen(cls, node: NodeAst, ns: VueCompNamespace):
        comp_ast = VueCompAst.transform(node.tag, node.attrs)

        # @computed
        def __html_tag_exit_gen_html():
            # v-if or v-show
            if_cond = comp_ast.v_if or comp_ast.v_show
            if if_cond and not if_cond.eval(ns):
                return ''

            # v-html
            if comp_ast.v_html:
                attr_chain = comp_ast.v_html
                # python expr
                return eval(attr_chain, {}, ns.to_py_eval_ns())

            # innerHtml
            inner = []
            for child in node.children:
                if callable(child):
                    inner.append(child())
                elif isinstance(child, widgets.HTML):
                    inner.append(child.value)
                else:
                    inner.append(child)
            inner_html = ' '.join(inner)

            attrs = [f"{k}='{v}'" for k, v in comp_ast.kwargs.items()]

            # v-bind:
            for attr, exp_ast in comp_ast.v_binds.items():
                attrs.append(f"{attr}='{to_raw(exp_ast.eval(ns))}'")

            html = f'''
                <{node.tag} {' '.join(attrs)}>
                  {inner_html}
                </{node.tag}>'''

            return html

        return __html_tag_exit_gen_html

    @classmethod
    def gen_from_fn(cls, fn):
        widget = widgets.HTML()

        @watch(fn, WatchOptions(immediate=True))
        def _update_html_widget_value(new_html, old_html, on_cleanup):
            widget.value = new_html

        return widget



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

from vuepy.compiler_sfc import sfc_compiler


def import_sfc(sfc_file, raw_content=False):
    """
    import Component from sfc_file or raw_content fo sfc

    :param sfc_file:
    :param raw_content:
    :return:
    """
    return sfc_compiler.compile(sfc_file, raw_content)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

from typing import Any
from typing import Callable
from typing import List

from vuepy import log as logging
from vuepy.reactivity import config

logger = logging.getLogger()

activeEffectScope: "EffectScope" = None


class EffectScope:
    def __init__(self, detached=False):
        self.active = True
        # self.effects: List["ReactiveEffect"] = []
        self.effects = []
        self.detached = detached
        self.cleanups: List[Callable[[], None]] = []
        self.scopes: List[EffectScope] = []
        self.index: int = None

        self.parent = activeEffectScope
        if not self.detached and activeEffectScope:
            self.index = len(activeEffectScope.scopes)
            activeEffectScope.scopes.append(self)

    def run(self, fn: Callable[[], Any]) -> Any:  # (fn: () => T): T
        if self.active:
            with self:
                return fn()
        elif config.__DEV__:
            logger.warn('cannot run an inactive effect scope.')

    def on(self):
        """
        This should only be called on non-detached scopes
        :return:
        """
        global activeEffectScope
        activeEffectScope = self

    def off(self):
        """
        This should only be called on non-detached scopes
        :return:
        """
        global activeEffectScope
        activeEffectScope = self.parent

    def __enter__(self):
        global activeEffectScope

        self.parent = activeEffectScope
        activeEffectScope = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        global activeEffectScope

        activeEffectScope = self.parent
        self.parent = None

    def clear(self):
        if not self.active:
            return

        for effect in self.effects:
            effect.stop()

        for cleanup in self.cleanups:
            cleanup()

        for scope in self.scopes:
            scope.stop(True)

    def stop(self, fromParent=False):
        if not self.active:
            return

        for effect in self.effects:
            effect.stop()

        for cleanup in self.cleanups:
            cleanup()

        for scope in self.scopes:
            scope.stop(True)

        if (not self.detached) and self.parent and (not fromParent):
            last = self.parent.scopes.pop() if self.parent.scopes else None
            if last and last is not self:
                if self.index < len(self.parent.scopes):
                    self.parent.scopes[self.index] = last
                last.index = self.index

        self.parent = None
        self.active = False


def effectScope(detached=False) -> EffectScope:
    return EffectScope(detached)


def recordEffectScope(effect, scope: "EffectScope" = None):
    if scope and scope.active:
        scope.effects.append(effect)


def getCurrentScope():
    return activeEffectScope


def onScopeDispose(fn: Callable[[], None]) -> None:
    if activeEffectScope:
        activeEffectScope.cleanups.append(fn)
    elif config.__DEV__:
        logger.warn("onScopeDispose() is called when there is no active effect scope to be associated with.")



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
import os
import shutil
import argparse
import re
from pathlib import Path
from string import Template

from vuepy.cli import log_error
from vuepy.cli import log_info

HERE = Path(__file__).absolute().parent


def is_valid_package_name(package_name):
    if len(package_name) > 214:
        return False
    if re.match(r'^[a-zA-Z0-9][a-zA-Z0-9_]*$', package_name) is None:
        return False
    return True


def create_project(args):
    project_name = input("✔ Project_name: ")
    template = args.template
    if not is_valid_package_name(project_name):
        log_error(f"Create project failed, Invalid package name: {project_name}")
        return

    proj_dir = Path(os.getcwd()) / project_name
    if proj_dir.exists():
        log_error(f"Create project failed, dir already exists: {proj_dir}")
        return
    else:
        os.makedirs(proj_dir)

    # 拷贝模板
    template_dir = HERE / f"template_{template}"
    if not template_dir.exists():
        log_error(f"Create project failed, Template not found: {template}")
        return

    ipywui_components_path = HERE.parent.parent / 'ipywui' / 'components'
    if not ipywui_components_path.exists():
        log_error(f"Create Project failed, ipywui not found: {ipywui_components_path}")
        return

    shutil.copytree(ipywui_components_path, proj_dir / 'ipywui')
    shutil.copytree(template_dir, proj_dir, dirs_exist_ok=True)

    with open(proj_dir / 'App.vue', 'r') as f:
        app_vue = f.read()
    app_vue = Template(app_vue).substitute({
        'js_stubs_path': HERE.parent / 'js_stubs',
        'ipywui_path': './ipywui',
    })
    with open(proj_dir / 'App.vue', 'w') as f:
        f.write(app_vue)

    log_info("")
    log_info(f"Scaffolding project in ./{project_name}")
    log_info("Done.")


def add_arg_parser(parser: argparse.ArgumentParser):
    # create <your project name> [--template vue]
    subparsers = parser.add_subparsers(help='create vue help')
    p = subparsers.add_parser('create', help='create a new app')
    # p.add_argument('project_name', type=str, help='The name of the project.')
    p.add_argument('--template', type=str, default='vue', help='The template to use.')
    p.set_defaults(func=create_project)


if __name__ == '__main__':
    # 创建项目
    parser = argparse.ArgumentParser(prog='PROG')
    add_arg_parser(parser)
    args = parser.parse_args()
    args.func(args)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import inspect
from typing import List

from ipywidgets import CallbackDispatcher

from vuepy import log
from vuepy.reactivity.ref import ref

logger = log.getLogger()


def defineProps(props: dict | list):
    """
    props = defineProps('p1')
    props.p1.value
    """
    frame = inspect.currentframe().f_back
    caller_args = get_caller_args(frame)
    init_props = caller_args[0] if caller_args else {}
    init_attrs = caller_args[1].get('attrs', {}) if len(caller_args) >= 2 else {}
    init_vals = {**init_props, **init_attrs}

    return DefineProps(props, init_vals)


class DefineProp:
    def __init__(self, prop_name: str, default=None):
        self.name = prop_name
        self._value = ref(default, debug_msg=f"prop:{prop_name}")

    @property
    def value(self):
        return self._value.value

    @value.setter
    def value(self, val):
        self._value.value = val


class DefineProps:
    def __init__(self, props: dict | list, init_vals: dict = None):
        self.prop_names = props
        self.props: List[DefineProp] = [
            DefineProp(name, init_vals.get(name)) for name in self.prop_names
        ]
        for prop in self.props:
            setattr(self, prop.name, prop)


class defineEmits:
    def __init__(self, events: List[str]):
        self.events = events
        self.events_to_cb_dispatcher: dict[str, CallbackDispatcher] = {}
        for event in self.events:
            self.add_event(event)

    def get_cb_dispatcher(self, event):
        return self.events_to_cb_dispatcher.get(event)

    def add_event(self, event):
        if event in self.events_to_cb_dispatcher:
            return
        self.events_to_cb_dispatcher[event] = CallbackDispatcher()

    def add_event_listener(self, event, callback, remove=False):
        cb_dispatcher = self.get_cb_dispatcher(event)
        if cb_dispatcher:
            cb_dispatcher.register_callback(callback, remove)

    def clear_events(self):
        self.events_to_cb_dispatcher = {}

    def __call__(self, event, *args, **kwargs):
        """$emit event.

        :param event:
        :param args: payload
        :param kwargs: payload
        :return:
        """
        handlers = self.events_to_cb_dispatcher.get(event)
        if not handlers:
            raise Exception(f"Event {event} not supported.")
        handlers(*args, **kwargs)


class defineModel:
    """
    count = defineModel("count")
    count.value += 1
    """
    # DEFAULT_KEY = 'modelValue'
    DEFAULT_KEY = 'value'

    def __init__(self, model_key: str | dict = DEFAULT_KEY):
        self.model_key = model_key
        self.prop = DefineProp(model_key)
        self.update_event = f'update:{self.model_key}'

    @property
    def value(self):
        return self.prop.value

    @value.setter
    def value(self, val):
        logger.debug("defineModel:%s set value %s to %s", self.model_key, self.prop.value, val)
        self.prop.value = val


def get_caller_args(frame):
    if not frame:
        return []

    caller_name = frame.f_code.co_name
    caller_func = frame.f_globals.get(caller_name)
    if not caller_func:
        logger.warn(f"can't get caller_func<{caller_name}>")
        return []

    argspec = inspect.getfullargspec(caller_func)
    if not argspec.args:
        logger.warn(f"get caller_func<{caller_name}> args is None")
        return []

    return [frame.f_locals.get(arg_name) for arg_name in argspec.args]



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

from vuepy.utils.common import Nil


class VueCompNamespace:
    def __init__(self, root, global_vars, local_vars=None):
        self.local_vars = local_vars or {}
        self.global_vars = global_vars
        self.root = root
        self.ns_list = [
            self.local_vars,
            self.global_vars,
        ]

    def to_py_eval_ns(self, tmp_vars=None):
        return {
            **self.global_vars,
            **self.local_vars,
            **(tmp_vars or {}),
        }

    @staticmethod
    def _getattr(obj, attr, default=Nil):
        if isinstance(obj, dict):
            return obj[attr] if default is Nil else obj.get(attr, default)
        else:
            return getattr(obj, attr) if default is Nil else getattr(obj, attr, default)

    @classmethod
    def get_by_attr_chain(cls, obj, attr_chain, default=Nil):
        for attr in attr_chain.split('.'):
            obj = cls._getattr(obj, attr, Nil)
            if obj is Nil:
                break
        else:
            return obj

        raise Exception(f"get attr {attr_chain} from ns failed")

    def get_obj_and_attr(self, attr_chain):
        obj_attr_tuple = attr_chain.rsplit('.', 1)
        for ns in self.ns_list:
            if not ns:
                continue

            if len(obj_attr_tuple) == 1:
                if attr_chain not in ns:
                    continue
                else:
                    return self.root, attr_chain

            obj_attr_chain, attr = obj_attr_tuple
            obj = self.get_by_attr_chain(ns, obj_attr_chain)
            if obj is not Nil:
                return obj, attr

        raise Exception(f"get attr {attr_chain} from ns failed")

    def getattr(self, attr_chain, default=Nil):
        return self.get_by_attr_chain(self.to_py_eval_ns(), attr_chain, default)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import asyncio
from collections.abc import Coroutine
from typing import Type

import ipywidgets as widgets

from vuepy import VueComponent
from vuepy import defineModel
from vuepy import watch
from vuepy.compiler_core.ast import NodeAst
from vuepy.compiler_core.ast import VueCompAst
from vuepy.compiler_core.utils import VueCompNamespace
from vuepy.compiler_dom.codegen import VueHtmlCompCodeGen
from vuepy.compiler_sfc.sfc_codegen import SFC
from vuepy.compiler_sfc.sfc_codegen import SFCType
from vuepy.compiler_sfc.codegen import logger
from vuepy.reactivity.reactive import to_raw
from vuepy.reactivity.ref import RefImpl
from vuepy.reactivity.watch import WatchOptions
from vuepy.utils.common import Nil
from vuepy.utils.common import has_changed

# todo move compiler_dom/codegen.py over here


class VueCompCodeGen:
    """
    h()
    """

    @classmethod
    def gen(
            cls,
            node: NodeAst,
            vm: 'VueComponent',
            ns: VueCompNamespace,
            app: "App"
    ):
        comp_ast = VueCompAst.transform(node.tag, node.attrs)
        # v-if
        if comp_ast.v_if:
            dummy = widgets.VBox()

            def _if_cond():
                return comp_ast.v_if.eval(ns)

            @watch(_if_cond, WatchOptions(immediate=True))
            def _change_v_if_widget(cond, old, on_cleanup):
                dummy.children = (cls._gen(comp_ast, node.children, vm, ns, app),) if cond else ()

            return dummy
        # v-show
        elif comp_ast.v_show:
            dummy = widgets.VBox()
            w = cls._gen(comp_ast, node.children, vm, ns, app)

            def _if_show():
                return comp_ast.v_show.eval(ns)

            @watch(_if_show, WatchOptions(immediate=True))
            def _show_widget(curr_show, old, on_cleanup):
                dummy.children = (w,) if curr_show else ()

            return dummy
        else:
            return cls._gen(comp_ast, node.children, vm, ns, app)

    @classmethod
    def _gen(
            cls,
            comp_ast: VueCompAst,
            children,
            vm: 'VueComponent',
            ns: VueCompNamespace,
            app: "App"
    ):
        component_cls: Type[VueComponent] = vm.component(comp_ast.tag)
        slots = {'default': []}
        for child in children or []:
            slot_name = getattr(child, 'v_slot', 'default')
            child = VueHtmlCompCodeGen.gen_from_fn(child) if callable(child) else child
            if slot_name == 'default':
                slots[slot_name].append(child)
            else:
                slots[slot_name] = child

        ctx = {
            'attrs': {**comp_ast.kwargs},
            'slots': slots,
            'emit': '',
        }
        # get v-bind init value
        props = {
            widget_attr: exp_ast.eval(ns)
            for widget_attr, exp_ast in comp_ast.v_binds.items()
        }
        # get v-model init value
        for _model_key, _attr_chain in comp_ast.v_model:
            # :bind, parent to child
            # widget_attr = component_cls.v_model_default  # VueCompTag.v_model(comp_ast.tag)
            if _model_key == defineModel.DEFAULT_KEY and hasattr(component_cls, 'v_model_default'):
                _widget_attr = getattr(component_cls, 'v_model_default')
            else:
                _widget_attr = _model_key

            props[_widget_attr] = to_raw(ns.getattr(_attr_chain))

        if isinstance(component_cls, SFCType):
            component = component_cls.gen(props, ctx, app)
        else:
            component = component_cls()
        widget = component.render(ctx, props, {})

        if comp_ast.tag == 'slot' and isinstance(vm, SFC):
            _slot_name = comp_ast.kwargs.get('name', 'default')
            _slot = vm._context.get('slots', {}).get(_slot_name)
            if _slot:
                widget.children = _slot

        # v-slot
        if comp_ast.v_slot:
            widget.v_slot = comp_ast.v_slot

        # v-bind:
        for widget_attr, exp_ast in comp_ast.v_binds.items():
            if widget_attr == 'style' and hasattr(widget, 'css_style'):
                widget_attr = 'css_style'

            def _get_v_bind_value(_exp_ast=exp_ast):
                return _exp_ast.eval(ns)

            @watch(_get_v_bind_value, WatchOptions(immediate=True))
            def _v_bind_update_vm_to_view(curr, old, on_cleanup, _widget_attr=widget_attr):
                _old_val = to_raw(getattr(widget, _widget_attr, Nil))
                curr = to_raw(curr)
                if has_changed(curr, _old_val):
                    setattr(widget, _widget_attr, curr)

        # v-model:define-model=bind
        for _model_key, attr_chain in comp_ast.v_model:
            # :bind, parent to child
            # widget_attr = component_cls.v_model_default  # VueCompTag.v_model(comp_ast.tag)
            widget_attr = _model_key
            if _model_key == defineModel.DEFAULT_KEY and hasattr(component_cls, 'v_model_default'):
                widget_attr = getattr(component_cls, 'v_model_default')

            def _get_v_model_value(_attr_chain=attr_chain):
                return to_raw(ns.getattr(_attr_chain))

            @watch(_get_v_model_value, WatchOptions(immediate=True))
            def _v_model_update_vm_to_view(curr, old, on_cleanup, _widget_attr=widget_attr):
                _old_val = to_raw(getattr(widget, _widget_attr, Nil))
                curr = to_raw(curr)
                if has_changed(curr, _old_val):
                    setattr(widget, _widget_attr, curr)

            # v-on, child to parent
            def listener(_vm, _obj, _attr):
                def _(change):
                    val = change['new'] if isinstance(change, dict) else change
                    setattr(_obj, _attr, val)
                    if isinstance(_obj, defineModel) and isinstance(_vm, SFC):
                        logger.debug("%s emit(%s, %s)", _vm, _obj.update_event, val)
                        getattr(_vm.root, SFC.EMIT_FN)(_obj.update_event, val)

                return _

            obj, attr = ns.get_obj_and_attr(attr_chain)
            cls.add_event_listener(
                widget,
                f'update:{widget_attr}',
                listener(vm, obj, attr),
            )

        # v-on
        for ev, func_ast in comp_ast.v_on.items():
            def _event_handle(*args, _func_ast=func_ast, **kwargs):
                return _func_ast.eval(ns, {'__vp_args': args, '__vp_kwargs': kwargs})

            cls.add_event_listener(widget, ev, _event_handle)

        if comp_ast.v_ref:
            _ref = ns.getattr(comp_ast.v_ref)
            # if not isinstance(_ref, VueRef):
            if not isinstance(_ref, RefImpl):
                logger.error(f'v-ref={comp_ast.v_ref} needs to be of type Ref.')
            else:
                _ref.value = component if isinstance(component_cls, SFCType) else widget

        return widget

    @staticmethod
    def add_event_listener(widget, event, listener):
        listener = wrap_callback(listener)
        if hasattr(widget, SFC.ADD_EVENT_LISTENER_FN):  # SFC
            getattr(widget, SFC.ADD_EVENT_LISTENER_FN)(event, listener)
        elif event.startswith('update:'):  # anywidget, ipyw
            attr = event.split(':', 1)[1]
            widget.observe(listener, names=attr)
        else:  # ipyw click
            getattr(widget, f"on_{event}")(listener)


def wrap_callback(callback):
    """
    Wraps a callback function to handle coroutines.
    If the callback returns a coroutine, it will be scheduled to run in the event loop.
    If the callback returns a non-coroutine, it will be returned immediately.
    """
    def _callback(*args, **kwargs):
        ret = callback(*args, **kwargs)
        if isinstance(ret, Coroutine):
            # for jupyter
            loop = asyncio.get_event_loop()
            if loop.is_running():
                return asyncio.create_task(ret)
            else:
                asyncio.run(ret)
        return ret
    return _callback



import time
from pathlib import Path

import anywidget
import traitlets


class MessageItem:
    def __init__(self, message_widget, msg_id):
        self.message_widget = message_widget
        self.msg_id = msg_id

    def close(self):
        self.message_widget.close_msg(self.msg_id)


class MessageWidget(anywidget.AnyWidget):
    _esm = Path(__file__).parent / 'index.js'
    _css = Path(__file__).parent / 'index.css'

    message_options = traitlets.Dict({
        "message": '',
        "type": "info",
        "show_close": False,
        "duration": 3000,
    }).tag(sync=True)
    close_msg_id = traitlets.Unicode().tag(sync=True)

    def close_msg(self, msg_id):
        self.close_msg_id = msg_id

    def _send_message(self, options):
        msg_id = f"msg-{str(time.time())}"
        options['msg_id'] = msg_id
        self.message_options = options
        return MessageItem(self, msg_id)

    def success(self, options):
        options['type'] = 'success'
        return self._send_message(options)

    def info(self, options):
        options['type'] = 'info'
        return self._send_message(options)

    def warning(self, options):
        options['type'] = 'warning'
        return self._send_message(options)

    def error(self, options):
        options['type'] = 'error'
        return self._send_message(options)


class MessageService:
    _instance = {}
    _is_inited = {}

    def __new__(cls, *args, **kwargs):
        app = kwargs.get("app_instance")
        if app not in cls._instance:
            cls._instance[app] = super().__new__(cls)
        return cls._instance[app]

    def __init__(self, app_instance=None):
        if app_instance not in MessageService._is_inited:
            self.app = app_instance
            self.widget = MessageWidget()
            MessageService._is_inited[app_instance] = True

    def __call__(self, options=None, app_instance=None):
        message_service = MessageService(app_instance=app_instance)
        if options:
            return message_service.widget._send_message(options)
        return message_service

    def success(self, options):
        return self.widget.success(options)

    def info(self, options):
        return self.widget.info(options)

    def warning(self, options):
        return self.widget.warning(options)

    def error(self, options):
        return self.widget.error(options)


Message = MessageService()



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import dataclasses
from dataclasses import field
from typing import List


@dataclasses.dataclass
class CompilerOptions:
    # Whitespace handling strategy
    whitespace = 'condense'  # preserve | condense
    delimiters: List[str] = field(default_factory=lambda: ['{{', '}}'])
    # Whether to keep comments in the templates AST
    comments: bool = False

    def is_custom_element(self, tag: str) -> bool:
        """
        Separate option for end users to extend the native elements list

        :param tag:
        :return:
        """
        pass



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import ast
import types

from _ast import Attribute
from typing import Any


# todo rename VueCompExpr
class VueCompExpr:
    def __init__(self, exp_ast, _vars=None, exp_str=''):
        self.vars = _vars if _vars else []
        self.exp_ast = exp_ast
        self.exp_str = exp_str

    def add_var(self, var):
        self.vars.append(var)

    @classmethod
    def parse(cls, expr_str) -> VueCompExpr:
        # return VueCompExprParser.parse(expr_str)
        expr_ast = ast.parse(expr_str, mode='eval')
        comp_expr_ast = VueCompExprTransformer().transformer(expr_ast)
        comp_expr_ast.exp_str = expr_str
        return comp_expr_ast

    @classmethod
    def compile(cls, exp_ast):
        if isinstance(exp_ast, types.CodeType):
            return exp_ast

        return compile(exp_ast, "<string>", "eval")

    def eval(self, ns: "VueCompNamespace", local_vars=None):
        # return vue_comp_expr_eval(self, ns, local_vars)
        code_obj = self.compile(self.exp_ast)
        return eval(code_obj, {"__builtin__": None}, ns.to_py_eval_ns(local_vars))


class VueCompExprTransformer(ast.NodeTransformer):
    def __init__(self):
        self.vars = []
        self._in_attr = False

    @classmethod
    def get_attr_chain(cls, node):
        if isinstance(node, ast.Name):
            return node.id
        elif isinstance(node, ast.Attribute):
            base = cls.get_attr_chain(node.value)
            return f"{base}.{node.attr}"

    def visit_Attribute(self, node: Attribute) -> Any:
        if self._in_attr:
            return node

        self._in_attr = True
        self.vars.append(self.get_attr_chain(node))

        self.generic_visit(node)
        self._in_attr = False
        return node

    def transformer(self, exp_ast: ast.Expression):
        if isinstance(exp_ast.body, ast.Name):
            self.vars = [exp_ast.body.id]
        else:
            self.visit(exp_ast)

        return VueCompExpr(exp_ast, self.vars)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
import enum


class LifecycleHooks(enum.Enum):
    BEFORE_CREATE = 'bc'
    CREATED = 'c'
    BEFORE_MOUNT = 'bm'
    MOUNTED = 'm'
    BEFORE_UPDATE = 'bu'
    UPDATED = 'u'
    BEFORE_UNMOUNT = 'bum'
    UNMOUNTED = 'um'
    DEACTIVATED = 'da'
    ACTIVATED = 'a'
    RENDER_TRIGGERED = 'rtg'
    RENDER_TRACKED = 'rtc'
    ERROR_CAPTURED = 'ec'
    SERVER_PREFETCH = 'sp'


class Hook:
    pass


class OnMounted:
    def __init__(self, callback):
        self.callback = callback


class OnBeforeMount:
    def __init__(self, callback):
        self.callback = callback


class OnUnmounted:
    def __init__(self, callback):
        self.callback = callback


class OnBeforeUnmount:
    def __init__(self, callback):
        self.callback = callback


def onMounted(callback):
    return OnMounted(callback)


def onBeforeMount(callback):
    return OnBeforeMount(callback)


def onUnmounted(callback):
    return OnUnmounted(callback)


def onBeforeUnmount(callback):
    return OnBeforeUnmount(callback)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import ast
import importlib.util
import types


class ScriptCompiler:
    @staticmethod
    def compile_script_block_bak(code_str, source_file_path):
        indent_code_str = '\n'.join([f"  {line}" for line in code_str.split('\n')])
        code = '\n'.join([
            "def setup(props, ctx, app):",
            indent_code_str,
            "\n",
            "  return locals()",
        ])

        module = types.ModuleType('tmp_module')
        module.__dict__['__file__'] = source_file_path
        exec(code, module.__dict__)
        return module.setup

    @staticmethod
    def compile_script_block(code_str, source_file_path):
        module = ast.parse(code_str)
        func_name = 'setup'
        func_ast = ast.FunctionDef(
            name=func_name,
            args=ast.arguments(
                posonlyargs=[],
                args=[
                    ast.arg(arg='props', annotation=None),
                    ast.arg(arg='ctx', annotation=None),
                    ast.arg(arg='app', annotation=None)
                ],
                vararg=None,
                kwonlyargs=[],
                kw_defaults=[],
                kwarg=None,
                defaults=[],
            ),
            body=module.body + [ast.parse('return locals()').body[0]],
            decorator_list=[],
            returns=None
        )

        module.body = [func_ast]
        ast.fix_missing_locations(module)
        code = compile(module, filename='<ast>', mode='exec')

        pymodule = types.ModuleType('tmp_module')
        pymodule.__dict__['__file__'] = source_file_path
        exec(code, pymodule.__dict__)
        return pymodule.setup

    @staticmethod
    def compile_script_src(dir_path, src):
        _script_path = dir_path.joinpath(src)
        if not _script_path.exists():
            raise ValueError(f"sfc script {_script_path} not exists.")

        _spec = importlib.util.spec_from_file_location('sfc', str(_script_path))
        _module = importlib.util.module_from_spec(_spec)
        _spec.loader.exec_module(_module)
        return getattr(_module, 'setup')



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import abc

import ipywidgets as widgets

from vuepy import log

logger = log.getLogger()


class Dom(widgets.VBox):
    """
    https://developer.mozilla.org/en-US/docs/Web/API/Document_Object_Model/Introduction
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def appendChild(self, el):
        self.children = self.children + (el,)
        return el

    def appendLeftChild(self, el):
        self.children = (el,) + self.children
        return el


class SetupContext:
    def __init__(self):
        self.attrs = {}
        self.slots = {}
        # 触发事件
        self.emit = {}
        # 暴露公共属性
        self.expose = {}


class VNode:
    def __init__(self, type_, props, children=None, key=''):
        self.type = type_
        self.props = props
        self.children = children
        self.key = key


class VueComponent(metaclass=abc.ABCMeta):
    _name = ''
    props = {}
    template = ''

    v_model_default = 'value'

    def __init__(self, setup_ret: dict = None, template: str = '', app: "App" = None):
        self.template = template
        self._data: dict = setup_ret
        self.app = app

    @classmethod
    def name(cls):
        # return (cls._name or cls.__name__).lower()
        return cls._name or cls.__name__

    def setup(self, props: dict = None, ctx=None, vm: 'App' = None):
        """

        :param props:
        :param ctx: attrs, slots, emit
        :param vm:
        :return:
        """
        pass

    def component(self, name: str):
        comp = self.to_ns().get(name)
        try:
            from vuepy.compiler_sfc.sfc_codegen import SFCType
            if comp and (isinstance(comp, SFCType) or issubclass(comp, VueComponent)):
                return comp
        except Exception as e:
            logger.debug(f"find component({name}) in Component `{self.name()}` failed, {e}.")
            pass

        comp = self.app.component(name)
        return comp

    def to_ns(self):
        return {
            **self.app.config.globalProperties.to_ns(),
            **self._data,
        }

    def render(self, ctx, props, setup_returned) -> VNode:
        """
        h(tag, props|attrs, children)

        :param ctx: attrs, slots, emit
        :param setup_returned:
        :param props:
        :return:
        """
        # if not self.template:
        #     return None
        # return vue_compiler_dom(self.template)
        pass



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import dataclasses
import pathlib
from typing import Any
from typing import Callable
from typing import List
from typing import Tuple

from ipywidgets import CallbackDispatcher

from vuepy.compiler_sfc.codegen import Dom
from vuepy.compiler_sfc.codegen import SetupContext
from vuepy.compiler_sfc.codegen import VNode
from vuepy.compiler_sfc.codegen import VueComponent
from vuepy.compiler_sfc.codegen import logger
from vuepy.reactivity.effect_scope import EffectScope
from vuepy.runtime.core.api_lifecycle import OnBeforeMount
from vuepy.runtime.core.api_lifecycle import OnMounted
from vuepy.runtime.core.api_setup_helpers import DefineProp
from vuepy.runtime.core.api_setup_helpers import DefineProps
from vuepy.runtime.core.api_setup_helpers import defineEmits
from vuepy.runtime.core.api_setup_helpers import defineModel
from vuepy.utils.common import Nil


class SFCWidgetContainer(Dom):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.on_mounted_dispatcher = CallbackDispatcher()
        self.on_before_mount_dispatcher = CallbackDispatcher()
        self.on_unmounted_dispatcher = CallbackDispatcher()
        self.on_before_unmount_dispatcher = CallbackDispatcher()


class SFC(VueComponent):
    ADD_EVENT_LISTENER_FN = '_s1_on'  # $on
    EMIT_FN = '_s1_emit'  # $emit

    def __init__(
            self,
            context: SetupContext | dict,
            props: dict,
            setup_ret: dict,
            template: str,
            app: "App",
            render: Callable[[SetupContext | dict, dict, dict], VNode] = None,
            file: str = None,
    ):
        super().__init__()
        self.app = app
        self.template = template
        self.setup_returned = setup_ret
        self._context = context
        self._props = props
        self._data: dict = setup_ret
        self._define_props = {}
        self._define_emits = {}
        self._render = render
        self.file = file
        self.scope: EffectScope = EffectScope(True)

        # setup define_x
        _emits, _on_fn = self._gen_widget_on_fn()
        _props_property = self._gen_widget_props_property()
        _model_update_events = self._gen_model_update_events()
        for _event in _model_update_events:
            _emits.add_event(_event)

        _sfc_container_cls = type('SfcContainer', (SFCWidgetContainer,), {
            **_props_property,
            self.ADD_EVENT_LISTENER_FN: _on_fn,
            self.EMIT_FN: _emits,
        })
        self.root: SFCWidgetContainer = _sfc_container_cls()
        self._init_static_props(context.get('attrs', {}))
        self._register_lifecycle_cb(self.root)

    def _init_static_props(self, attrs):
        if not self.define_props:
            return

        _, _props = self.define_props
        for prop_name in _props.prop_names:
            val = attrs.get(prop_name, Nil)
            if val is Nil:
                continue
            setattr(self.root, prop_name, val)

    def _register_lifecycle_cb(self, container: SFCWidgetContainer):
        for _, cb in self._get_var_by_type(OnBeforeMount):
            container.on_before_mount_dispatcher.register_callback(cb.callback)

        for _, cb in self._get_var_by_type(OnMounted):
            container.on_mounted_dispatcher.register_callback(cb.callback)

    def setup(self, props: dict = None, ctx=None, vm: 'App' = None):
        return self._data

    def _gen_model_update_events(self):
        return [_model.update_event for (_, _model) in self.define_models]

    def _gen_widget_props_property(self):
        def _get(prop: DefineProp):
            def _(this):
                return prop.value

            return _

        def _set(prop: DefineProp):
            def _(this, val):
                prop.value = val

            return _

        props = [model.prop for _, model in self.define_models]
        if self.define_props:
            props.extend(self.define_props[1].props)

        return {
            prop.name: property(_get(prop), _set(prop))
            for prop in props
        }

    def _gen_widget_on_fn(self):
        emits = self.define_emits[1] if self.define_emits else defineEmits([])

        def _on_fn(this, event, callback, remove=False):
            emits.add_event_listener(event, callback, remove)

        return emits, _on_fn

    @property
    def define_props(self) -> (str, DefineProps):
        ret = self._get_var_by_type(DefineProps, limit=1)
        return ret[0] if ret else []

    @property
    def define_emits(self) -> (str, defineEmits):
        ret = self._get_var_by_type(defineEmits, limit=1)
        return ret[0] if ret else []

    @property
    def define_models(self) -> List[(str, defineModel)]:
        return self._get_var_by_type(defineModel)

    def _get_var_by_type(self, define_type, limit: int = float('inf')) -> List[Tuple[str, Any]]:
        ret = []
        count = 0
        for var_name, var in self.setup_returned.items():
            if count >= limit:
                break
            if isinstance(var, define_type):
                ret.append((var_name, var))
                count += 1
        return ret

    def __repr__(self):
        return f"{self.__class__.__name__}<{pathlib.Path(self.file).name}> at {id(self)}"

    def render(self, ctx: SetupContext = None, props: dict = None, setup_returned: dict = None) -> VNode:
        logger.info(f"🔥 Rerender {self}")
        self.scope.clear()

        # beforeMount
        self.root.on_before_mount_dispatcher()

        # compile and render
        with self.scope:
            if self._render:
                dom = self._render(self._context, self._props, self.setup_returned)
            else:
                from vuepy.compiler_sfc.template_compiler import DomCompiler
                dom = DomCompiler(self, self.app).compile(self.template)

        self.root.children = [dom]

        # mounted
        self.root.on_mounted_dispatcher()

        return self.root


@dataclasses.dataclass
class SFCType:
    # (props: dict, context: SetupContext, app: App) -> dict | Callable[[], h]:
    setup: Callable[[dict, SetupContext | dict, 'App'], dict | Callable] = None
    template: str = ''
    # (self, ctx, props, setup_returned) -> VNode:
    render: Callable[[SetupContext | dict, dict, dict], VNode] = None
    _file: str = ''

    def gen(self, props: dict, context: SetupContext | dict, app: "App") -> "SFC":
        setup_ret = self.setup(props, context, app) if self.setup else {}
        return SFC(context, props, setup_ret, self.template, app, self.render, self._file)



#!coding: utf-8
# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
import ipywui
from ipywui.core import IPywidgetsComponent
from ipywui.core import is_float
from ipywui.core import wui
from ipywui.widgets import ClipboardWidget
from ipywui.widgets import DialogWidget
from ipywui.widgets import DisplayViewer
from ipywui.widgets import MarkdownViewerWidget
from vuepy import log as logging

logger = logging.getLogger()


@wui.register()
class AppLayout(IPywidgetsComponent):
    def render(self, ctx, props, setup_returned):
        slots = ctx.get('slots', {})
        attrs = ctx.get('attrs', {})
        return ipywui.widgets.AppLayout(**slots, **props, **attrs)


@wui.register()
class VBox(IPywidgetsComponent):
    def _render(self, ctx, attrs, props, params, setup_returned):
        slots = ctx.get('slots', {})
        return ipywui.widgets.VBox(children=slots.get('default', []), **props, **attrs, **params)


@wui.register('template')
class Template(VBox):
    pass


@wui.register('slot')
class Slot(VBox):
    pass


@wui.register()
class Box(IPywidgetsComponent):
    def _render(self, ctx, attrs, props, params, setup_returned):
        slots = ctx.get('slots', {})
        return ipywui.widgets.VBox(children=slots.get('default', []), **props, **attrs, **params)


@wui.register()
class HBox(IPywidgetsComponent):
    def _render(self, ctx, attrs, props, params, setup_returned):
        slots = ctx.get('slots', {})
        return ipywui.widgets.HBox(children=slots.get('default', []), **props, **attrs, **params)


@wui.register()
class AccordionItem(IPywidgetsComponent):

    def _render(self, ctx, attrs, props, params, setup_returned):
        slots = ctx.get('slots', {})
        title = attrs.pop('title', '-')
        widget = ipywui.widgets.VBox(children=slots.get('default', []), **props, **attrs, **params)
        widget.title = title
        return widget


@wui.register()
class Accordion(IPywidgetsComponent):
    v_model_default = 'selected_index'

    def _render(self, ctx, attrs, props, params, setup_returned):
        slots = ctx.get('slots', {})
        children = slots.get('default', [])
        titles = [child.title for child in children]
        return ipywui.widgets.Accordion(children=children, titles=titles, **props, **attrs, **params)


@wui.register()
class Button(IPywidgetsComponent):
    v_model_default = 'description'

    PARAMS_STORE_TRUE = [
        ('loading', False),
        ('disabled', False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        slots = ctx.get('slots', {}).get('default', [])
        slot_label = slots[0] if slots else None
        if slot_label:
            attrs[self.v_model_default] = slot_label.value

        # params > attrs > props
        merged_props = {**props, **attrs, **params}
        widget = ipywui.widgets.Button(**merged_props)

        if slot_label:
            def _update_html_widget_value(change):
                val = change['new'] if isinstance(change, dict) else change
                setattr(widget, self.v_model_default, val)

            slot_label.observe(_update_html_widget_value, 'value')

        return widget


@wui.register()
class Checkbox(IPywidgetsComponent):
    v_model_default = 'value'

    def render(self, ctx, props, setup_returned):
        attrs = ctx.get('attrs', {})
        self.update_style(attrs)
        self.update_style(props)
        return ipywui.widgets.Checkbox(**props, **attrs)


@wui.register()
class ColorPicker(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ('disabled', False),
        ('concise', False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.ColorPicker(**props, **attrs, **params)


@wui.register()
class Combobox(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ('disabled', False),
        ("continuous_update", False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.Combobox(**props, **attrs, **params)


@wui.register()
class Controller(IPywidgetsComponent):
    v_model_default = 'value'

    def render(self, ctx, props, setup_returned):
        attrs = ctx.get('attrs', {})
        return ipywui.widgets.Controller(**props, **attrs)


@wui.register()
class Clipboard(IPywidgetsComponent):
    def render(self, ctx, props, setup_returned):
        attrs = ctx.get('attrs', {})
        slots = ctx.get('slots', {})
        return ClipboardWidget(children=slots.get('default', []), **props, **attrs)


@wui.register()
class Col(IPywidgetsComponent):
    def _render(self, ctx, attrs, props, params, setup_returned):
        slots = ctx.get('slots', {})
        widget = ipywui.widgets.VBox(children=slots.get('default', []), **props, **attrs)

        widget.span = attrs.pop('span', 24)
        widget.offset = attrs.pop('offset', 0)
        return widget


@wui.register()
class DatePicker(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ("disabled", False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.DatePicker(**props, **attrs, **params)


@wui.register()
class DateTimePicker(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ("disabled", False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        # todo widgets.NaiveDatetimePicker
        return ipywui.widgets.DateTimePicker(**props, **attrs, **params)


@wui.register()
class Dialog(IPywidgetsComponent):
    """
    仅支持放在最外层的VBOX中
    """
    v_model_default = 'value'

    def render(self, ctx, props, setup_returned):
        slots = ctx.get('slots', {})
        slot_body = slots.get('default', slots.get('body', []))
        slot_footer = slots.get('footer', [])
        body = slot_body if isinstance(slot_body, list) else [slot_body]
        footer = slot_footer if isinstance(slot_footer, list) else [slot_footer]

        attrs = ctx.get('attrs', {})
        width = attrs.pop("layout", {}).pop("width", '50%')
        params = {
            'width': width,
        }
        widget = DialogWidget(body=body, footer=footer, **props, **attrs, **params)
        return widget


@wui.register()
class Display(IPywidgetsComponent):
    PARAMS_STORE_TRUE = [
        ("multi_thread", False),
    ]

    def render(self, ctx, props, setup_returned):
        attrs = ctx.get('attrs', {})
        obj = props.pop('obj', '-')
        widget = DisplayViewer(obj, **props, **attrs)
        return widget


@wui.register()
class Dropdown(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ("disabled", False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.Dropdown(**props, **attrs, **params)


@wui.register()
class FileUpload(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ("multiple", False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.FileUpload(**props, **attrs, **params)


@wui.register()
class HTMLMath(IPywidgetsComponent):
    v_model_default = 'value'

    def render(self, ctx, props, setup_returned):
        attrs = ctx.get('attrs', {})
        return ipywui.widgets.HTMLMath(**props, **attrs)


@wui.register()
class Image(IPywidgetsComponent):
    v_model_default = 'value'

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.Image(**props, **attrs, **params)


@wui.register()
class Input(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ("disabled", False),
        ("continuous_update", False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        _type = attrs.get('type', props.get('type', 'text'))
        if _type == 'password':
            cls = ipywui.widgets.Password
        elif _type == 'textarea':
            cls = ipywui.widgets.Textarea
        else:
            cls = ipywui.widgets.Text

        return cls(**props, **attrs, **params)


@wui.register()
class InputNumber(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ("disabled", False),
    ]
    CAN_DEFAULT = 0
    CAN_FLOAT = 1
    CAN_BOUNDED = 2
    # key CAN_FLOAT | CAN_BOUNDED -> cls
    IMPLS = {
        CAN_FLOAT | CAN_BOUNDED: ipywui.widgets.BoundedFloatText,
        CAN_BOUNDED: ipywui.widgets.BoundedIntText,
        CAN_FLOAT: ipywui.widgets.FloatText,
        CAN_DEFAULT: ipywui.widgets.IntText,
    }

    def _render(self, ctx, attrs, props, params, setup_returned):
        attrs = ctx.get('attrs', {})

        _value = props.get(self.v_model_default)
        _step = props.get('step')
        _min = props.get('min')
        _max = props.get('max')
        is_float_value = any(is_float(v) for v in (_value, _step, _min, _max))
        is_bounded_value = any(val is not None for val in (_min, _max))

        can_float = self.CAN_FLOAT if is_float_value else self.CAN_DEFAULT
        can_bounded = self.CAN_BOUNDED if is_bounded_value else self.CAN_DEFAULT

        cls = self.IMPLS.get(can_float | can_bounded, None)
        if cls is None:
            msg = f"Invalid input number type: CAN_FLOAT | CAN_BOUNDED = {can_float} | {can_bounded}"
            raise ValueError(msg)

        return cls(**props, **attrs, **params)


@wui.register()
class Label(IPywidgetsComponent):
    v_model_default = 'value'

    def _render(self, ctx, attrs, props, params, setup_returned):
        slots = ctx.get('slots', {}).get('default', [])
        slot_label = slots[0] if slots else None
        if slot_label:
            attrs[self.v_model_default] = slot_label.value

        # params > attrs > props
        merged_props = {**props, **attrs, **params}
        widget = ipywui.widgets.Label(**merged_props)

        if slot_label:
            def _update_html_widget_value(change):
                val = change['new'] if isinstance(change, dict) else change
                setattr(widget, self.v_model_default, val)

            slot_label.observe(_update_html_widget_value, 'value')

        return widget


@wui.register()
class MarkdownViewer(IPywidgetsComponent):
    v_model_default = 'value'

    def render(self, ctx, props, setup_returned):
        attrs = ctx.get('attrs', {})
        return MarkdownViewerWidget(**props, **attrs)


@wui.register()
class Play(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ('disabled', False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.Play(**props, **attrs, **params)


@wui.register()
class Progress(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ('vertical', False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        params["orientation"] = 'vertical' if params.get("vertical") else 'horizontal'

        return ipywui.widgets.FloatProgress(**props, **attrs, **params)


@wui.register()
class RadioButtons(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ('disabled', False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.RadioButtons(**props, **attrs, **params)


@wui.register()
class Row(IPywidgetsComponent):
    N_ROWS = 1
    N_COLS = 24

    def _render(self, ctx, attrs, props, params, setup_returned):
        slots = ctx.get('slots', {})
        cols = slots.get('default', [])

        justify_content = attrs.pop('justify', props.pop('justify', None))
        if justify_content:
            params['justify_content'] = justify_content

        align_items = attrs.pop('align', props.pop('align', None))
        if align_items:
            params['align_items'] = align_items

        grid_gap = attrs.pop('gutter', props.pop('gutter', None))
        if grid_gap:
            params['grid_gap'] = grid_gap

        widget = ipywui.widgets.GridspecLayout(self.N_ROWS, self.N_COLS, **props, **attrs, **params)
        idx = 0
        for col in cols:
            idx += col.offset
            widget[0, slice(idx, idx + col.span)] = col
            idx += col.span
        return widget


@wui.register()
class Select(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ('disabled', False),
        ('multiple', False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        multiple = params.pop('multiple', False)
        if multiple:
            cls = ipywui.widgets.SelectMultiple
        else:
            cls = ipywui.widgets.Select

        return cls(**props, **attrs, **params)


@wui.register()
class SelectColors(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ('unique', False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.ColorsInput(**props, **attrs, **params)


@wui.register()
class SelectNumbers(IPywidgetsComponent):
    v_model_default = 'value'

    def _render(self, ctx, attrs, props, params, setup_returned):
        attrs = ctx.get('attrs', {})
        data_type = attrs.pop('data_type', props.pop('data_type', 'float'))
        value = props.get(self.v_model_default, [])
        if data_type == 'float' or any(is_float(n) for n in value):
            cls = ipywui.widgets.FloatsInput
        else:
            cls = ipywui.widgets.IntsInput

        return cls(**props, **attrs)


@wui.register()
class SelectTags(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ('unique', False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.TagsInput(**props, **attrs, **params)


@wui.register()
class Slider(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ("disabled", False),
        ("continuous_update", False),
        ('vertical', False),
        ('range', False),
    ]

    CAN_DEFAULT = 0
    CAN_RANGE = 1
    CAN_SELECTION = 2
    CAN_FLOAT = 4
    # key CAN_RANGE | CAN_SELECTION | CAN_FLOAT -> cls
    IMPLS = {
        CAN_RANGE | CAN_SELECTION: ipywui.widgets.SelectionRangeSlider,
        CAN_RANGE | CAN_FLOAT: ipywui.widgets.FloatRangeSlider,
        CAN_RANGE: ipywui.widgets.IntRangeSlider,
        CAN_SELECTION: ipywui.widgets.SelectionSlider,
        CAN_FLOAT: ipywui.widgets.FloatSlider,
        CAN_DEFAULT: ipywui.widgets.IntSlider,
    }

    def _render(self, ctx, attrs, props, params, setup_returned):
        params["orientation"] = 'vertical' if params.get("vertical") else 'horizontal'
        attrs = ctx.get('attrs', {})

        is_range = params.pop('range', False)
        is_selection = props.get('options', attrs.get('options', None)) is not None
        if is_range:
            _values = props.get(self.v_model_default, [])
        else:
            _values = [props.get(self.v_model_default)]
        _min = props.get('min')
        _max = props.get('max')
        _step = props.get('step')
        is_float_value = any(is_float(v) for v in (*_values, _min, _max, _step))

        can_range = self.CAN_RANGE if is_range else self.CAN_DEFAULT
        can_selection = self.CAN_SELECTION if is_selection else self.CAN_DEFAULT
        can_float = self.CAN_FLOAT if is_float_value else self.CAN_DEFAULT

        cls = self.IMPLS.get(can_range | can_selection | can_float, None)
        if cls is None:
            msg = f"Invalid slider type: CAN_RANGE | CAN_SELECTION | CAN_FLOAT = {can_range} | {can_selection} | {can_float}"
            raise ValueError(msg)

        return cls(**props, **attrs, **params)


@wui.register()
class Stack(IPywidgetsComponent):
    v_model_default = 'selected_label'

    def _render(self, ctx, attrs, props, params, setup_returned):
        slots = ctx.get('slots', {})
        children = slots.get('default', [])
        labels = [child.label for child in children]
        return ipywui.widgets.Stack(children=children, labels=labels, **props, **attrs, **params)


@wui.register()
class StackItem(IPywidgetsComponent):
    def _render(self, ctx, attrs, props, params, setup_returned):
        slots = ctx.get('slots', {})
        label = props.pop('label', attrs.pop('label', '-'))
        widget = ipywui.widgets.VBox(children=slots.get('default', []), **props, **attrs, **params)
        widget.label = label
        return widget


@wui.register()
class TabPane(IPywidgetsComponent):
    def _render(self, ctx, attrs, props, params, setup_returned):
        slots = ctx.get('slots', {})
        title = props.pop('title', attrs.pop('title', '-'))
        widget = ipywui.widgets.VBox(children=slots.get('default', []), **props, **attrs, **params)
        widget.title = title
        return widget


@wui.register()
class Tabs(IPywidgetsComponent):
    v_model_default = 'selected_index'

    def _render(self, ctx, attrs, props, params, setup_returned):
        slots = ctx.get('slots', {})
        children = slots.get('default', [])
        titles = [child.title for child in children]
        return ipywui.widgets.Tab(children=children, titles=titles, **props, **attrs, **params)


@wui.register()
class TimePicker(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ("disabled", False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.TimePicker(**props, **attrs, **params)


@wui.register()
class ToggleButton(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ('disabled', False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.ToggleButton(**props, **attrs, **params)


@wui.register()
class ToggleButtons(IPywidgetsComponent):
    v_model_default = 'value'

    PARAMS_STORE_TRUE = [
        ('disabled', False),
    ]

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.ToggleButtons(**props, **attrs, **params)


@wui.register()
class Valid(IPywidgetsComponent):
    v_model_default = 'value'

    def _render(self, ctx, attrs, props, params, setup_returned):
        return ipywui.widgets.Valid(readout='', **props, **attrs, **params)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import dataclasses

from vuepy.log import getLogger

logger = getLogger()


def has_changed(value, old) -> bool:
    try:
        return bool(value != old)
    except Exception as e:
        logger.warn(f"Run has_changed failed, {e}")
        return True


def gen_hash_key(target):
    try:
        hash(target)
    except Exception:
        return f"id(target)={id(target)})"
    return target


class Nil:
    pass


@dataclasses.dataclass
class Record:
    def to_ns(self):
        return self.__dict__



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from typing import Type
from typing import Union

from vuepy.compiler_sfc.codegen import VueComponent
from vuepy.compiler_sfc.sfc_codegen import SFCType
from vuepy.utils.factory import FactoryMeta


class VuepyAppStore(metaclass=FactoryMeta):
    """
    VuepyAppStore stores all the Vuepy App: Union[Type[VueComponent], SFCType])).
    """

    @classmethod
    def register(cls, app_name, app: Union[Type[VueComponent], SFCType]):
        """
        Register a Vuepy Component.

        :param app_name: The name of the App.
        :param app: The App to register.
        :return:
        """
        FactoryMeta.register(cls, app_name)(app)
        return app

    @classmethod
    def get(cls, app_name) -> Union[Type[VueComponent], SFCType]:
        """
        Get a Vuepy App by name.

        :param app_name: The name of the App.
        :return: The App.
        """
        return cls.get_all_registry().get(app_name)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import functools
import traceback
from html.parser import HTMLParser
from typing import List
from typing import Tuple

import ipywidgets as widgets

from vuepy import log
from vuepy.compiler_core.ast import NodeAst
from vuepy.compiler_core.ast import NodeAstType
from vuepy.compiler_core.ast import VForAst
from vuepy.compiler_core.ast import VForNodeAst
from vuepy.compiler_core.ast import VForScopes
from vuepy.compiler_core.ast import VueCompAst
from vuepy.compiler_core.errors import CompilerSyntaxError
from vuepy.compiler_core.utils import VueCompNamespace
from vuepy.compiler_dom.codegen import VueHtmlCompCodeGen
from vuepy.compiler_dom.codegen import VueHtmlTemplateRender
from vuepy.compiler_dom.codegen import v_for_stack_to_iter
from vuepy.compiler_sfc.template_codegen import VueCompCodeGen
from vuepy.compiler_sfc.codegen import VueComponent
from vuepy.reactivity.watch import watch
from vuepy.runtime.core.api_create_app import App
from vuepy.utils.common import Nil

logger = log.getLogger()


class DomCompiler(HTMLParser):
    """
    @vue/compiler-dom
    """

    def __init__(self, vm: VueComponent, app: 'App'):
        super().__init__()
        self.vm = vm
        self.app = app
        self.widgets = NodeAst('dummy', children=[])
        self.widgets_by_id = {}
        self.parent_node_stack: List[NodeAst | VForNodeAst] = []
        self.v_for_stack: List[VForAst] = []
        self.v_if_stack: List[str] = []
        self._tag = self.widgets.tag

        self._compile_traceback_info = ''

    def _get_element_by_id(self, el_id):
        return self.widgets_by_id.get(el_id)

    @property
    def is_in_for_stmt(self):
        return bool(self.v_for_stack)

    def _process_directive(self):
        pass

    def _gen_widget(self, node: NodeAst, for_scope: VForScopes = None):
        local_vars = for_scope and for_scope.to_ns()
        ns = VueCompNamespace(self.vm.to_ns(), self.vm.to_ns(), local_vars)
        if self.vm.component(node.tag):
            widget = VueCompCodeGen.gen(node, self.vm, ns, self.app)
        else:
            widget = VueHtmlCompCodeGen.gen(node, ns)

        return widget

    def _get_raw_tag(self, tag):
        row, col = self.getpos()
        row_idx = row - 1
        col_idx = col + 1
        return self.html_lines[row_idx][col_idx: col_idx + len(tag)]

    def _to_camel_case_tag(self, tag):
        words = tag.split('-')
        if len(words) == 1:
            return tag
        return ''.join(w.title() for w in words)

    def handle_starttag(self, case_insensitive_tag, attrs):
        def _gen_node_ast(tag, attrs, for_processed: bool = False, idxs: Tuple[int] = None,
                          vars: dict = None, v_for: VForAst = None) -> NodeAst:
            idxs = idxs or []
            if self.parent_node_stack:
                _parent = self.parent_node_stack[-1]
                _parent = _parent.children if isinstance(_parent, VForNodeAst) else _parent
            else:
                _parent = self.widgets

            for idx in (idxs[:-1] if for_processed else idxs):
                _parent = _parent[idx]

            _for_scopes = VForScopes(idxs=idxs, vars=vars) if idxs and vars else None
            _node = NodeAst(
                tag=tag, attrs=attrs, parent=_parent, children=[], plain=False, v_for=v_for,
                v_for_scopes=_for_scopes, for_processed=for_processed,
                type=node_type
            )
            return _node

        raw_tag = self._get_raw_tag(case_insensitive_tag)
        tag = self._to_camel_case_tag(raw_tag)
        node_type = NodeAstType.WIDGET if self.vm.component(tag) else NodeAstType.HTML
        self._tag = tag
        attrs = dict(attrs)
        self._transformer_node_attrs(attrs)

        v_for_expr = attrs.pop(VueCompAst.V_FOR, None)
        v_for_ast = VForAst.parse(v_for_expr)
        for_processed = False
        if v_for_ast:
            for_processed = True
            self.v_for_stack.append(v_for_ast)

            ns = VueCompNamespace(self.vm.to_ns(), self.vm.to_ns())
            attr_chain = v_for_ast.iter

            def __track_list_change():
                # track list replacements
                # python expr
                obj_iter = eval(attr_chain, {}, ns.to_py_eval_ns())
                # track changes in the list itself, such as append, pop...
                return id(obj_iter), [id(item) for item in obj_iter]

            @watch(__track_list_change)
            def __track_list_change_rerender(new, old, on_cleanup):
                self.vm.render()

        if self.v_for_stack:
            nodes = v_for_stack_to_iter(
                self.v_for_stack,
                functools.partial(_gen_node_ast, tag, attrs, for_processed),
                self.vm._data
            )
            node = VForNodeAst(tag, attrs, children=nodes, for_processed=for_processed,
                               type=node_type)
        else:
            node = _gen_node_ast(tag, attrs, for_processed)
        self.parent_node_stack.append(node)

    def _transformer_node_attrs(self, attrs: dict) -> None:
        prev_v_if_expr = self.v_if_stack.pop() if self.v_if_stack else ''

        _v_if_expr = attrs.get(VueCompAst.V_IF)
        if _v_if_expr:
            return

        v_else_if_expr = attrs.pop(VueCompAst.V_ELSE_IF, '')
        if v_else_if_expr:
            if not prev_v_if_expr:
                _err_msg = f"v-else-if={v_else_if_expr} error, there is no corresponding v-if"
                raise CompilerSyntaxError(_err_msg)

            v_else_if_expr = f"(not ({prev_v_if_expr})) and ({v_else_if_expr})"
            attrs[VueCompAst.V_IF] = v_else_if_expr
            return

        v_else_expr = attrs.pop(VueCompAst.V_ELSE, Nil)
        if v_else_expr is not Nil:
            if not prev_v_if_expr:
                _err_msg = f"v-else error, there is no corresponding v-if"
                raise CompilerSyntaxError(_err_msg)

            v_else_expr = f"not ({prev_v_if_expr})"
            attrs[VueCompAst.V_IF] = v_else_expr
            return

    def handle_data(self, data: str) -> None:
        if not self.parent_node_stack:
            return

        if not data.strip():
            return

        tag = self._tag

        def _gen_text(node: NodeAst, should_render: bool):
            if isinstance(node, VForNodeAst):
                for _node in node.children_flat:
                    _gen_text(_node, should_render)
                return

            ns = VueCompNamespace(self.vm.to_ns(), self.vm.to_ns(),
                                  node.v_for_scopes and node.v_for_scopes.to_ns())
            i = -1
            if node.v_for_scopes:
                i = node.v_for_scopes.idxs[-1]

            def __handle_data_gen_html(_ns=ns, _i=i, _should_render=should_render):
                if _should_render:
                    logger.debug("for_stmt_gen_html index=%s <%s> tmpl=`%s`", _i, tag, data)
                    return VueHtmlTemplateRender.render(data, _ns, _i)
                else:
                    return data

            if node.type == NodeAstType.HTML:
                node.add_child(__handle_data_gen_html)
            else:
                node.add_child(VueHtmlCompCodeGen.gen_from_fn(__handle_data_gen_html))

        # ./_gen_text()

        should_render = VueHtmlTemplateRender.should_render(data)
        parent = self.parent_node_stack[-1]
        _gen_text(parent, should_render)

    def handle_endtag(self, tag):
        def _gen_element(_node: NodeAst):
            if isinstance(_node, VForNodeAst):
                for _node in _node.children_flat:
                    _gen_element(_node)
                return

            widget = self._gen_widget(_node, _node.v_for_scopes)
            widget._vtag = widget.__class__.__name__
            # not v-for
            if not _node.v_for_scopes:
                if callable(widget):
                    widget = VueHtmlCompCodeGen.gen_from_fn(widget)
                    widget._vtag = tag
            # curr v-for end
            elif _node.for_processed:
                if callable(widget):
                    widget = VueHtmlCompCodeGen.gen_from_fn(widget)
                    widget._vtag = tag
                widget._vtag = f"{widget._vtag} v-for: {_node.v_for_scopes.idxs}"
            # v-for in process
            else:
                pass
            _node.parent.add_child(widget)

        # ./_gen_element

        node = self.parent_node_stack.pop()
        _gen_element(node)
        if node.for_processed:
            self.v_for_stack.pop()

        v_if_expr = node.attrs.get(VueCompAst.V_IF)
        if v_if_expr:
            self.v_if_stack.append(v_if_expr)

    def _compile_traceback(self, e: Exception):
        s = ''
        ident = ''
        # already generated node
        for node in self.widgets.children:
            tag, attr = self._get_node_debug_tag(node)
            s += f"{ident}<{tag}{attr}>...</{tag}>\n"

        for level, _node in enumerate(self.parent_node_stack):
            ident = '  ' * level
            tag, attr = self._get_node_debug_tag(_node)
            s += f"{ident}<{tag}{attr}>\n"
            if isinstance(_node, VForNodeAst):
                continue

            pre_ident = ident
            for child in _node.children:
                # compile _node in process
                if isinstance(child, NodeAst):
                    continue
                ident = f"  {pre_ident}"
                tag, attr = self._get_node_debug_tag(child)
                s += f"{ident}<{tag}{attr}>...</{tag}>\n"

        traceback_info = ''.join(traceback.format_tb(e.__traceback__))
        if self.parse_starttag.__name__ in traceback_info:
            s += f"  {ident}<{self._tag}> start"
        elif self.parse_endtag.__name__ in traceback_info:
            s += f"{ident}</{self._tag}> end"
        else:
            s += f"{ident}</{self._tag}> inner"

        s = f"{s} <----- compile failed, {e}\n"
        return s

    def _get_node_debug_tag(self, node):
        """

        :param node: NodeAst or widget
        :return:
        """
        if isinstance(node, NodeAst):
            tag, attr = node.tag, ''
        else:
            vtag = getattr(node, '_vtag', node.__class__.__name__).split(' ', 1)
            tag, attr = vtag if len(vtag) == 2 else (vtag[0], '')
            attr = f" {attr}" if attr else ""

        if isinstance(node, VForNodeAst):
            attr += f' v-for'

        return tag, attr

    def compile(self, html):
        self.html_lines = [line for line in html.splitlines()]
        try:
            self.feed(html)
        except Exception as e:
            self._compile_traceback_info = self._compile_traceback(e)
            raise e

        if len(self.widgets.children) == 1:
            return self.widgets.children[0]
        return widgets.VBox(self.widgets.children)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

from vuepy.compiler_sfc.script_compiler import ScriptCompiler
from vuepy.compiler_sfc.sfc_codegen import SFCType
from vuepy.compiler_sfc import sfc_parser


def compile(sfc_file, raw_content=False) -> SFCType:
    """
    Compile a sfc file.

    :param sfc_file: The path to the sfc file.
    :param raw_content: If True, the sfc file is parsed as a string.
    :return: SFCType
    """
    sfc_meta = sfc_parser.parse(sfc_file, raw_content)
    if sfc_meta.script_src:
        setup_fn = ScriptCompiler.compile_script_src(sfc_meta.file.parent, sfc_meta.script_src)
    elif sfc_meta.script_py:
        setup_fn = ScriptCompiler.compile_script_block(
            sfc_meta.script_py, str(sfc_meta.file.absolute()))
    else:
        setup_fn = lambda *args: {}

    return SFCType(**{
        'setup': setup_fn,
        'template': sfc_meta.template,
        '_file': sfc_meta.file,
    })



# coding: utf-8
"""Setup script for vuepy"""

import os.path

from setuptools import find_packages
from setuptools import setup

# The directory containing this file
HERE = os.path.realpath(os.path.dirname(__file__))

# The text of the README file
with open(os.path.join(HERE, "README.md")) as fid:
    README = fid.read()

# This call to setup() does all the work
setup(
    name="vuepy-core",
    version="0.1.10",
    description="Vue.py is a progressive, incrementally-adoptable Python framework for building UI on the Jupyter Notebook.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://www.vuepy.org",
    keywords=["vue", "Web UI", "MVVM"],
    author="vuepy.org",
    author_email="leilux.dev@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
    ],
    python_requires='>=3.7',
    packages=find_packages('src', exclude=('*widgets*', 'vuepy2')),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=[
        "ipywidgets",
        "markdown",
        "anywidget",
    ],
    entry_points={"console_scripts": ["vuepy=vuepy.__main__:main"]},
)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import enum
from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import List

from vuepy import log
from vuepy.reactivity import config
from vuepy.reactivity.constant import IterateKey
from vuepy.reactivity.dep import Dep
from vuepy.reactivity.dep import createDep
from vuepy.reactivity.effect_scope import EffectScope
from vuepy.reactivity.effect_scope import getCurrentScope
from vuepy.reactivity.effect_scope import recordEffectScope
from vuepy.utils.common import gen_hash_key

logger = log.getLogger()


class DepStore:
    def __init__(self):
        # self._store = weakref.WeakKeyDictionary()
        self._store: dict[str, Dep] = {}  # { target_id: Dep }

    def get_or_create(self, target) -> Dep:
        key = gen_hash_key(target)
        if key not in self._store:
            self._store[key] = createDep(key=key)
        return self._store[key]

    def get(self, target) -> Dep:
        return self._store.get(gen_hash_key(target))

    def len(self):
        return len(self._store)

    def clear(self):
        return self._store.clear()

    def __contains__(self, target) -> bool:
        return gen_hash_key(target) in self._store


class ReactiveDepStore:
    """
    dep for reactive
    {
        id(target): DepStore({
            id(key1): Dep(effect..),
            id(key2): Dep(effect..),
            ...
        }),
        ...
    }
    """

    def __init__(self):
        self._store = {}

    def get_or_create(self, target) -> DepStore:
        key = gen_hash_key(target)
        if key not in self._store:
            self._store[key] = DepStore()
        return self._store[key]

    def get(self, target) -> DepStore:
        return self._store.get(gen_hash_key(target))

    def len(self):
        return len(self._store)

    def delete(self, target):
        self._store.pop(gen_hash_key(target), None)

    def clear(self):
        self._store.clear()

    def __contains__(self, target) -> bool:
        return gen_hash_key(target) in self._store


targetMap = ReactiveDepStore()  # { target: { key: Dep } }

shouldTrack = True
activeEffect: "ReactiveEffect" | None = None
trackOpBit = 1
effectTrackDepth = 1


class IgnoreTracking:
    def __init__(self):
        self.shouldTrack = None

    def __enter__(self):
        global shouldTrack
        self.shouldTrack = shouldTrack
        shouldTrack = False

    def __exit__(self, exc_type, exc_val, exc_tb):
        global shouldTrack
        shouldTrack = self.shouldTrack
        self.shouldTrack = False


class TrackOpTypes(enum.Enum):
    GET = 'get'
    HAS = 'has'
    ITER = 'iter'


class TriggerOpTypes(enum.Enum):
    SET = 'set'
    ADD = 'add'
    DELETE = 'delete'
    CLEAR = 'clear'
    ITER = 'iter'


EffectScheduler = Callable[..., Any]


class ReactiveEffect:
    def __init__(self, fn, scheduler: EffectScheduler = None, scope: EffectScope = None):
        self.fn = fn
        self.scheduler = scheduler
        self.scope = scope
        self.active = True
        self.parent: ReactiveEffect = None
        self.computed = None
        self.allow_recurse = False
        self.deferStop = False
        self.onStop = None
        self.onTrack = None
        self.onTrigger = None
        self.deps = []

        self.scope = getCurrentScope()
        recordEffectScope(self, self.scope)

    def __repr__(self):
        name = f"{self.fn.__name__} of {self.__class__.__name__} at {id(self)}"
        return f"{name}"

    def run(self):
        if not self.active:
            return self.fn()

        global activeEffect, trackOpBit, shouldTrack, effectTrackDepth
        parent = activeEffect
        last_should_track = shouldTrack
        while parent:
            if parent is self:
                return
            parent = parent.parent

        try:
            self.parent = activeEffect
            activeEffect = self
            shouldTrack = True

            effectTrackDepth += 1
            trackOpBit = 1 << effectTrackDepth
            # todo effectTrackDepth <= maxMarkerBits
            initDepMarkers(self.deps)
            return self.fn()
        finally:
            finalizeDepMarkers(self)

            effectTrackDepth -= 1
            trackOpBit = 1 << effectTrackDepth
            activeEffect = self.parent
            shouldTrack = last_should_track
            self.parent = None
            if self.deferStop:
                self.stop()

    def stop(self):
        if activeEffect is self:
            self.deferStop = True
        elif self.active:
            cleanupEffect(self)
            if self.onStop:
                self.onStop()
            self.active = False


# todo move to reactiveeffect
def cleanupEffect(effect: ReactiveEffect):
    for dep in effect.deps:
        dep.delete(effect)
    effect.deps = []


def track(target, track_type: TrackOpTypes, key, msg=""):
    if not (shouldTrack and activeEffect):
        return

    eventInfo = config.__DEV__ and {
        'effect': activeEffect,
        'target': target,
        'type': track_type,
        'key': key,
        'msg': msg,
    }
    if isinstance(target, list) and track_type == TrackOpTypes.ITER:
        key = IterateKey.get_key(target)
    dep = targetMap.get_or_create(target).get_or_create(key)
    trackEffects(dep, eventInfo)


def trackEffects(dep: Dep, debugger_info: DebuggerEventExtraInfo = None):
    should_track = False
    if not newTracked(dep):
        dep.n |= trackOpBit
        should_track = not wasTracked(dep)
    # else:
    #     should_track = activeEffect not in dep
    if not should_track:
        return

    dep.add(activeEffect)
    activeEffect.deps.append(dep)
    if config.__DEV__ and activeEffect.onTrack:
        activeEffect.onTrack(debugger_info)


def trigger(
        target,
        trigger_type: TriggerOpTypes,
        key=None,
        new=None,
        old=None,
        old_target=None,
        msg="",
):
    depsMap = targetMap.get(target)
    if not depsMap:
        return

    logger.debug(f"trigger %s.%s %s", target, key, msg)

    deps: List[Dep] = []
    if trigger_type in (TriggerOpTypes.CLEAR, TriggerOpTypes.ITER):
        deps.extend(depsMap.values())
    else:
        if key in depsMap:
            deps.append(depsMap.get(key))

        iter_key = IterateKey.get_key(target)

        if trigger_type == TriggerOpTypes.ADD:
            deps.append(depsMap.get(iter_key))
        elif trigger_type == TriggerOpTypes.DELETE:
            deps.append(depsMap.get(iter_key))
        elif trigger_type == TriggerOpTypes.SET:
            deps.append(depsMap.get(iter_key))

    eventInfo = config.__DEV__ and {
        'target': target,
        'type': trigger_type,
        'key': key,
        'newValue': new,
        'oldValue': old,
        'oldTarget': old_target,
        'msg': msg,
    }
    effects = []
    for dep in deps:
        if not dep:
            continue
        effects.extend(dep)

    triggerEffects(effects, eventInfo)


def triggerEffects(
        dep: Dep | List[ReactiveEffect],
        debuggerEventExtraInfo: DebuggerEventExtraInfo = None
):
    effects = dep if isinstance(dep, list) else list(dep)
    logger.debug("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^")
    logger.debug("🚀trigger effects%s len(%s) start", id(effects), len(effects))
    for effect in list(effects):
        if effect.computed:
            trigggerEffect(effect, debuggerEventExtraInfo)

    for effect in list(effects):
        if not effect.computed:
            trigggerEffect(effect, debuggerEventExtraInfo)
    logger.debug(f"effects%s len(%s) end🚀", id(effects), len(effects))
    logger.debug("vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv")


def trigggerEffect(effect: ReactiveEffect, debuggerEventExtraInfo=None):
    if effect is not activeEffect or effect.allow_recurse:
        if config.__DEV__ and effect.onTrigger:
            debuggerEventExtraInfo['effect'] = effect
            effect.onTrigger(debuggerEventExtraInfo)

        if effect.scheduler:
            effect.scheduler()
        else:
            effect.run()


@dataclass
class DebuggerEventExtraInfo:
    target: Any
    type: TrackOpTypes | TriggerOpTypes
    key: Any
    newValue: Any = None
    oldValue: Any = None
    oldTarget: dict[any, any] | set[any] = None


@dataclass
class DebuggerEvent(DebuggerEventExtraInfo):
    effect: ReactiveEffect = None


@dataclass
class DebuggerOptions:
    onTrack: Callable[[DebuggerEvent], None] = None
    onTrigger: Callable[[DebuggerEvent], None] = None


@dataclass
class ReactiveEffectOptions(DebuggerOptions):
    lazy: bool = False
    scheduler: EffectScheduler = None
    scope: EffectScope = None
    allowRecurse: bool = False
    onStop: Callable[[], None] = None

    def extend(self, effect: ReactiveEffect):
        effect.onTrack = self.onTrack
        effect.onTrigger = self.onTrigger
        effect.scheduler = self.scheduler
        effect.allow_recurse = self.allowRecurse
        effect.onStop = self.onStop
        if self.scope:
            recordEffectScope(effect, self.scope)


@dataclass
class ReactiveEffectRunner:
    effect: ReactiveEffect = None

    def __call__(self):
        pass


def effect_impl(fn, options: ReactiveEffectOptions = None):
    _effect = ReactiveEffect(fn)
    if options:
        options.extend(_effect)

    if not options or not options.lazy:
        _effect.run()

    runner = ReactiveEffectRunner()
    runner.effect = _effect
    return runner


def effect(fn_or_options=None, options: ReactiveEffectOptions = None):
    is_decorator_no_arg = (fn_or_options, options) == (None, None)
    is_decorator_with_arg = isinstance(fn_or_options, ReactiveEffectOptions)
    # @effect() or @effect(options)
    if is_decorator_with_arg or is_decorator_no_arg:
        def wrap(fn):
            return effect_impl(fn, fn_or_options)

        return wrap
    # @effect or effect(fn)
    else:
        return effect_impl(fn_or_options, options)


MUTABLE_TYPES = (dict, list, set)


def _can_reactive(target) -> bool:
    return isinstance(target, MUTABLE_TYPES)


def wasTracked(dep: Dep):
    return (dep.w & trackOpBit) > 0


def newTracked(dep: Dep):
    return (dep.n & trackOpBit) > 0


def initDepMarkers(deps: List[Dep]):
    for dep in deps:
        dep.w |= trackOpBit


def finalizeDepMarkers(effect):
    # todo 可优化
    ptr = 0
    dep_len = len(effect.deps)
    for i, dep in enumerate(effect.deps):
        if wasTracked(dep) and (not newTracked(dep)):
            dep.delete(effect)
        else:
            if i != ptr:
                effect.deps[ptr] = dep
            ptr += 1

        dep.w &= ~trackOpBit
        dep.n &= ~trackOpBit

    if ptr != dep_len:
        effect.deps = effect.deps[:ptr]



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations


class Dep(set):
    def __init__(self, *args, **kwargs):
        self.key = kwargs.pop('key', None)
        self.w = kwargs.pop('w', 0)  # wasTracked
        self.n = kwargs.pop('n', 0)  # newTracked
        super().__init__(*args, **kwargs)

    def add(self, effect) -> None:
        super().add(effect)

    def delete(self, effect) -> None:
        try:
            self.remove(effect)
        except Exception:
            pass


def createDep(effects=None, key=None) -> Dep:
    return Dep(effects or "", key=key)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import dataclasses
import enum
from typing import Any
from typing import Callable
from typing import List
from typing import Union

from vuepy.reactivity import config
from vuepy.reactivity.computed import ComputedRefImpl
from vuepy.reactivity.constant import ErrorCodes
from vuepy.reactivity.effect import DebuggerOptions
from vuepy.reactivity.effect import ReactiveEffect
from vuepy.reactivity.reactive import ReactiveProxy
from vuepy.reactivity.reactive import ReactiveFlags
from vuepy.reactivity.reactive import isShallow
from vuepy.reactivity.reactive import is_reactive
from vuepy.reactivity.ref import RefImpl
from vuepy.reactivity.ref import isRef
from vuepy.utils.common import Nil
from vuepy.utils.common import gen_hash_key
from vuepy.utils.common import has_changed

OnCleanUp = Callable[[Callable[[], None]], None]  # (cleanupFn: () => void) => void
WatchEffect = Callable[[OnCleanUp], None]  # (onCleanup: OnCleanup) => void
WatchSource = Union[RefImpl, ComputedRefImpl, Callable[[], Any], ReactiveProxy]
# (val: Any, oldVal: Any, onCleanup: OnCleanup) => any
WatchCallback = Callable[[Any, Any, OnCleanUp], Any]
MultiWatchSources = List[WatchSource]
WatchStopHandle = Callable[[], None]  # () => void


class FlushEnum(enum.Enum):
    PRE = 'pre'
    POST = 'post'
    SYNC = 'sync'


@dataclasses.dataclass
class WatchOptionsBase(DebuggerOptions):
    flush: FlushEnum = FlushEnum.PRE


@dataclasses.dataclass
class WatchOptions(WatchOptionsBase):
    immediate: bool = False
    deep: bool = False


def watchEffect(
        effect_or_options: WatchEffect | WatchOptions = None,
        debuggerOptions: DebuggerOptions = None
) -> WatchStopHandle:
    is_decorator_no_arg = (effect_or_options, debuggerOptions) == (None, None)
    is_decorator_with_arg = isinstance(effect_or_options, DebuggerOptions)
    # @watchEffect() or @watchEffect(debuggerOptions)
    if is_decorator_no_arg or is_decorator_with_arg:
        def wrap(fn):
            return watchEffect_impl(fn, effect_or_options)

        return wrap
    # @watchEffect or watchEffect()
    else:
        return watchEffect_impl(effect_or_options, debuggerOptions)


def watchEffect_impl(watch_effect: WatchEffect, options: WatchOptionsBase = None) -> WatchStopHandle:
    options = options or WatchOptionsBase()
    flush = options.flush
    onTrack = options.onTrack
    onTrigger = options.onTrigger

    instance = None

    cleanup = None

    def onCleanup(fn):
        nonlocal cleanup

        def _cleanup():
            nonlocal cleanup
            callWithErrorHandling(fn, instance, ErrorCodes.WATCH_CLEANUP)
            cleanup = None
            effect.onStop = None

        cleanup = _cleanup
        effect.onStop = _cleanup

    def getter():
        if instance and instance.isUnmounted:
            return
        cleanup and cleanup()
        return callWithErrorHandling(watch_effect, instance, ErrorCodes.WATCH_CALLBACK, [onCleanup])

    def job():
        if not effect.active:
            return
        effect.run()

    job.allowRecurse = False
    if flush == FlushEnum.SYNC:
        scheduler = job
    elif flush == FlushEnum.POST:
        scheduler = job
    else:
        scheduler = job

    effect = ReactiveEffect(getter, scheduler)
    if config.__DEV__:
        effect.onTrack = onTrack
        effect.onTrigger = onTrigger

    # initial run
    if flush == FlushEnum.POST:
        # queuePostRenderEffect()
        pass
    else:
        effect.run()

    def unwatch():
        effect.stop()
        if instance and instance.scope:
            instance.scope.effects.pop(effect)

    return unwatch


def watch(
        source: WatchSource | MultiWatchSources,
        cb_or_options: WatchCallback | WatchOptions = None,
        options: WatchOptions = None
) -> WatchStopHandle:
    is_decorator_src = (cb_or_options, options) == (None, None)
    is_decorator_src_options = isinstance(cb_or_options, WatchOptions)
    # @watch(src) or @watch(src, options)
    if is_decorator_src or is_decorator_src_options:
        def wrap(fn):
            return doWatch(source, fn, cb_or_options)

        return wrap
    # watch(src, cb) or watch(src, cb, options)
    else:
        return doWatch(source, cb_or_options, options)


def doWatch(
        source: WatchSource | List[WatchSource] | WatchEffect,
        cb: WatchCallback,
        watch_options: WatchOptions = None,
) -> WatchStopHandle:
    watch_options = watch_options or WatchOptions()
    immediate = watch_options.immediate
    deep = watch_options.deep
    flush = watch_options.flush
    onTrack = watch_options.onTrack
    onTrigger = watch_options.onTrigger

    # todo
    instance = None
    getter = None
    forceTrigger = False
    isMultiSource = False

    if isRef(source):
        getter = lambda: source.value
        forceTrigger = isShallow(source)
    elif is_reactive(source):
        getter = lambda: source
        deep = True
    elif isinstance(source, list):
        isMultiSource = True
        forceTrigger = any(is_reactive(s) or isShallow(s) for s in source)

        def getter():
            _ret = []
            for s in source:
                if isRef(s):
                    _ret.append(s.value)
                elif is_reactive(s):
                    _ret.append(traverse(s))
                elif callable(s):
                    _ret.append(callWithErrorHandling(s, instance, ErrorCodes.WATCH_GETTER))
                else:
                    config.__DEV__ and print(f"Invalid watch source: {source}.")
            return _ret

    elif callable(source):
        getter = lambda: callWithErrorHandling(source, instance, ErrorCodes.WATCH_GETTER)
    else:
        getter = lambda: None
        config.__DEV__ and print(f"Invalid watch source: {source}.")

    if cb and deep:
        baseGetter = getter
        getter = lambda: traverse(baseGetter())

    cleanup = None

    def onCleanup(fn):
        nonlocal cleanup

        def _cleanup():
            nonlocal cleanup
            callWithErrorHandling(fn, instance, ErrorCodes.WATCH_CLEANUP)
            cleanup = None
            effect.onStop = None

        cleanup = _cleanup
        effect.onStop = _cleanup

    oldValue = [Nil] * len(source) if isMultiSource else Nil

    def job():
        nonlocal oldValue
        if not effect.active:
            return

        newValue = effect.run()
        if isMultiSource:
            had_change = any(has_changed(n, o) for (n, o) in zip(newValue, oldValue))
        else:
            had_change = has_changed(newValue, oldValue)

        if deep or forceTrigger or had_change:
            cleanup and cleanup()

            if oldValue is Nil:
                _old_val = None
            else:
                _old_val = [] if isMultiSource and (oldValue[0] is Nil) else oldValue
            _args = [newValue, _old_val, onCleanup]
            callWithErrorHandling(cb, instance, ErrorCodes.WATCH_CALLBACK, _args)
            oldValue = newValue

    job.allowRecurse = True
    if flush == FlushEnum.SYNC:
        scheduler = job
    elif flush == FlushEnum.POST:
        scheduler = job
    else:
        scheduler = job

    effect = ReactiveEffect(getter, scheduler)
    if config.__DEV__:
        effect.onTrack = onTrack
        effect.onTrigger = onTrigger

    # initial run
    if immediate:
        job()
    else:
        oldValue = effect.run()

    def unwatch():
        effect.stop()
        if instance and instance.scope:
            instance.scope.effects.pop(effect)

    return unwatch


@dataclasses.dataclass
class SchedulerJob:
    id: int = None
    pre: bool = False
    active: bool = False
    computed: bool = False
    allowRecurse: bool = False
    # ownerInstance: ComponentInternalInstance = None


def callWithErrorHandling(fn, instance, type, args: List = None):
    res = None
    try:
        res = fn(*args) if args else fn()
    except Exception as err:
        print(f"call {fn}({ args }) {type} failed, {err}")
        raise err
    return res


def traverse(value, seen: set = None):
    if not is_reactive(value) or getattr(value, ReactiveFlags.SKIP.value, False):
        return value

    seen = seen or set()
    value_hash = gen_hash_key(value)
    if value_hash in seen:
        return value

    seen.add(value_hash)
    if isRef(value):
        traverse(value.value, seen)
    elif isinstance(value, list):
        for val in value:
            traverse(val, seen)
    elif isinstance(value, dict):
        for val in value.values():
            traverse(val, seen)
    # todo
    # elif isPlaninObject(value):
    #     for key in value:
    #         traverse(getattr(value, key, None), seen)

    return value
